function [u,varargout] = ...
    iter_thresh(f,w,p,K,method,varargin)
%function [u,Psihist,reshist,Dhist,stephist,exitflag] = 
%    iter_thresh(f,w,p,K,method);
% Calculates the minimizer of
% (*)  Psi(u) = ||Ku - f||^2/2 + Phi(u)
% where Phi(u) = sum_k w_k |u_k|^p. For 0<p<=2 it can employ the
% forward-backward splitting method (e.g. analyzed in [DDD04] for
% 1<=p<=2 and in [BL09] for 0<=p<=1)  while for 1<=p<=2 it can only use
% the algorithms proposed in [BL08].
%
% Input:
%  f: from (*)
%  w: positive sequence of weights
%  p: exponent of penalty term
%  K: operator given either as a matrix or as a function handle
%  method: 'hard' minimizes by "iterated hard thresholding" accroding [BL08]
%          'soft' minimizes by forward-backward splitting from [DDD04], [BL09]
%          'combined' alternates between soft and hard thresholding
% Optional input:
%  'KT': The transpose operator of K as a function handle. (If K is given 
%        as a function handle, this has to be given.)
%  'Initialization': initial value for the iteration (default: 0)
%  'maxIter': maximal number of iterations (default: 300)
%  'Verbose': no output on screen (0, default), lots of output (1)
%  'Tol_r': tolerance for stopping because the residual norm, i.e. if
%              r = ||u - S_(w,p)(u - K'(Ku - f))||
%       falls below this threshold: r < Tol_r (default: 1e-6)
%       (Here S_(w,p) is the soft-thresholding function defined in [DDD04]
%       in eq. (2.4)-(2.6), i.e. r measures how accurate the fixed-point
%       equation is fulfilled.)
%  'Tol_D':tolerance for stopping because the value D from [BL08] falls
%       below this threshold: D < Tol_D (default: 1e-8)
%
% Output:
%  u: Result of the minimization
% Optional output:
%  Psihist: History of the values of Psi
%  reshist: History of the residual norm
%       r = ||u - S_w,p(u- K'(Ku-f))||
%  stephist: History of the stepsizes
%  Dhist: History of the estimate D on the distance to the minimal value,
%       see [BL08]
%  exitflag: true if iteration exited because of convergence, false if 
%            iteration runned till maxIter
%
% Note that the 'soft' option only converges for ||K||<sqrt(2)
% (recommended: ||K||<1). If this is not the case, a rescaling
%   K <- cK,
%   f <- cf
% with a small c solves the problem (e.g. c = 1/norm(K)).
%
% Example:
%  N = 300;
%  x = linspace(0,1,N)';
%  u = 40*(x>.3 & x<.31) - 80*(x>.55 & x<.56) + 52*(x>.62 & x<.63) - 12*(x>.87 & x<.88);
%  K = tril(ones(N))/N;
%  f = K*u + randn(size(u))*.05;
%  p = 1;
%  w = .01;
%  [u_s,Psi_s,D_s] = iter_thresh(f,w,p,K,'soft','Verbose',1,'maxIter',1000);
%  [u_h,Psi_h,D_h] = iter_thresh(f,w,p,K,'hard','Verbose',1,'maxIter',1000);
%  [u_c,Psi_c,D_c] = iter_thresh(f,w,p,K,'combined','Verbose',1,'maxIter',1000);
%  subplot(2,2,1)
%  plot(x,u), title('True solution u')
%  subplot(2,2,2)
%  plot(x,f), title('Noisy data f')
%  subplot(2,2,3)
%  plot(Psi_s),hold on, plot(Psi_h,'g'), plot(Psi_c,'r'), title('Objective value')
%  legend('Soft thresholding','Hard thresholding','Combination')
%  subplot(2,2,4)
%  plot(x,u_s,x,u_h,x,u_c), title('Reconstructed solution')
%  legend('Soft thresholding','Hard thresholding','Combination')
%
%References:
% [DDD04] An iterative thresholding algorithm for linear inverse problem
% with a sparsity constraint, I. Daubechies, M. Defrise, C. De Mol, Comm.
% Pure Appl. Math., 57(11):1413-1457, 2004
% [BL08] Iterated hard shrinkage for minimization problems with sparsity
% constraints, K. Bredies, D. Lorenz, SIAM Journal on Scientific Computing,
% 30(2):657-683, 2008
% [BL09]Minimization of non-smooth, non-convex functionals by iterative
% thresholding, K. Bredies, . Lorenz, Submitted, April 2009
%
% Dirk Lorenz, 10.07.2009, d.lorenz@tu-braunschweig.de


%Set default values for initial parameters 
maxIter = 300;
verbose = 0;
Tol_r = 1e-6;
Tol_D = 1e-8;
exitflag = false;
u0 = [];
KT = [];

% Read the optional parameters
if (rem(length(varargin),2)==1)
  error('Optional parameters should always go by pairs');
else
  for i=1:2:(length(varargin)-1)
    switch varargin{i}
     case 'Initialization'
       u0 = varargin{i+1};
     case 'maxIter'
         maxIter = varargin{i+1};
     case 'Verbose'
       verbose = varargin{i+1};
     case 'Tol_r'
       Tol_r = varargin{i+1};
     case 'Tol_D'
       Tol_D = varargin{i+1};
     case 'KT'
       KT = varargin{i+1};
     otherwise
      % Hmmm, something wrong with the parameter string
      error(['Unrecognized option: ''' varargin{i} '''']);
    end;
  end;
end

if isa(K, 'function_handle')
    if isempty(KT)
        error('KT must be given if K is a function handle')
    end
else
    KT = @(x) K'*x;
    K = @(x) K*x;
end


%Initialize histories
Psihist = zeros(1,maxIter);
stephist = zeros(1,maxIter);
Dhist = zeros(1,maxIter);
reshist = zeros(1,maxIter);

w0 = min(w(:));
if p > 0
    S0 = (sum(f(:).^2)/2/w0)^(1/p);
else
    S0 = inf;
end

if S0 == 0;
    u = 0*KT(f);
    varargout(1) = {[]};
    varargout(2) = {[]};
    varargout(3) = {[]};
    varargout(4) = {[]};
    varargout(5) = {true};
    return
end

if strcmp(method,'soft')
    S0 = 1e10;
end

%Precalculate K'f
Ktf = KT(f);

if isempty(u0)
    u0 = 0*Ktf;
end

u = u0;
    
if verbose
    fprintf(' Iteration | Psi       | r         | D         | s        |\n')
end

for iter = 1:maxIter
   Ku = K(u);
   KKu = KT(Ku);
   res = norm(u(:) - Sp(u(:) - (KKu(:) - Ktf(:)),p,w(:)));
   reshist(iter) = res;
   if strcmp(method,'hard') || (strcmp(method,'combined') && mod(iter,2) == 0)
        Psihist(iter) = sum((Ku(:) -f(:)).^2)/2 + sum(w(:).*phip(u(:),p,S0));
        S0 = (Psihist(iter)/w0).^(1/p);
        %Descent direction according to (5)
        v = Hp((-KKu + Ktf)./w,p,S0);
        %Step-size according to (6)
        D = penaltyfunctional(u,w,p,S0) - penaltyfunctional(v,w,p,S0) + ...
                    (KKu(:)-Ktf(:))' * (u(:)-v(:));
        Dhist(iter) = D;
        Kv = K(v);
        s = min(1,D/sum((Kv(:)-Ku(:)).^2));
        if D < Tol_D || sum((Kv(:)-Ku(:)).^2)==0 || res < Tol_r 
            exitflag = true;
            break
        end

        stephist(iter) = s;
        u = u + s*(v-u);
        
   elseif strcmp(method,'soft')|| (strcmp(method,'combined') && mod(iter,2) == 1)
        Psihist(iter) = sum((Ku(:) -f(:)).^2)/2 + sum(w(:).*abs(u(:)).^p);
        s = 1;
        stephist(iter) = s;
        v = Sp(u - s*(KKu - Ktf),p,s*w);
        D = penaltyfunctional(u,w,p,S0) - penaltyfunctional(v,w,p,S0) + ...
                    (KKu(:)-Ktf(:))' * (u(:)-v(:));
        Dhist(iter) = D;
        if (D < Tol_D) || (res < Tol_r)
            exitflag = true;
            break
        end
        u = v;
        
   end
   if verbose
       fprintf('      % 4d | %3.3e | %3.3e | %3.3e | %3.3e |\n',...
           iter,Psihist(iter),reshist(iter),Dhist(iter),stephist(iter))
   end
    
end

if exitflag
    fprintf(['\n iter_thresh with method "' method '" converged after a total number of %d iterations\n'],iter)
    fprintf(' and an objective value of %0.3e.\n', Psihist(iter))
else
    fprintf(['\n iter_thresh with method "' method '" exited because the maximal number\n'],iter)
    fprintf(' of iterations was reached. The final objective value is %0.3e.\n', Psihist(iter))
end

varargout(1) = {Psihist(1:iter)};
varargout(2) = {reshist(1:iter)};
varargout(3) = {Dhist(1:iter)};
varargout(4) = {stephist(1:iter)};
varargout(5) = {exitflag};

%%%%%%%%%%%%%%%%
% Subfunctions %
%%%%%%%%%%%%%%%%

function y = Hp(x,p,S0)
%function y = Hp(x,p,S0);
% Implements the hard shrinkage function H_p corresponding to the penalty
% phip.

if p>2 || p<1
    error('The hard-method only works for 1 <= p <= 2!')
elseif p>1 && p<=2
    y = sign(x).*(abs(x)/p.*(abs(x)<= p*S0^(p-1))).^(1/(p-1)) + ...
        S0^(2-p)*x/p.*(abs(x)>p*S0^(p-1));
elseif p==1
    y = 0.*(abs(x)<= 1) + ...
        S0*x.*(abs(x)>1);
end

function y = phip(x,p,S0)
%function y = phip(x,p,S0)
% Implements the penalty function phi_p which is a |x|^p for |x| < S0 and
% quadratic for |x| > S0. phip is of class C^1.

if p>2
    error('The soft-method only works for p <= 2!')
elseif p>=1 && p<=2
    y = abs(x).^p.*(abs(x)<= S0) + ...
        p/2/S0^(2-p)*(x.^2 + (2/p-1)*S0^2).*(abs(x)>S0);
elseif 0 <= p && p <=1
    y = abs(x).^p;
end

function v = Sp(u, p, lambda)
% V = Sp(U,P,LAMBDA)
%    For each element of u >= 0, the function returns the solution v
%    of the equation u = v + lambda*p*v^(p-1).
%    In the case p=1 and p=2, the explicit solution is known while
%    in the other cases, Newton iteration is performed with suitable
%    initial conditions.
[n,m] = size(u);
u = u(:);
su = sign(u);
u = abs(u);

if (p == 1)
    v = max(0, u - lambda(:));
elseif (p == 0)
    v = u.*(abs(u)>sqrt(2*lambda));
elseif (p == 2)
    v = u./(1 + 2*lambda);
elseif (p > 0) && (p < 1)
    thresh = (2-p)/(2-2*p)*(2*lambda*(1-p)).^(1/(2-p));
    v = u;
	for k = 1:length(u)
        if abs(u(k)) <= thresh  
            v(k) = 0;
        else
            % Newton's method
            z = u(k);
            l = lambda(k);
            for j=1:4
                z = z - (z+l*p*sign(z)*abs(z).^(p-1)-u(k))./(1+l.*p*(p-1)*abs(z)^(p-2));
            end
            v(k) = z;
        end
	end
else
    iter = 20;
    v = (u./(lambda*p*(2-p))).^(1/(p-1))/2;
    v = min(v, u./(1 + p*lambda));
    for i=1:iter
        v = (u - lambda.*(2-p)*p.*abs(v).^(p-1))./(1 + lambda.*(p-1)*p.*abs(v).^(p-2));
    end  
end
v = v.*su;
v = reshape(v,n,m);

function y = penaltyfunctional(u,w,p,S0)
%function y = penaltyfunctional(u,w,p,S0)
% Implements the penalty functional
%  Phip(u) = sum_k w_k phi_p(u_k).

y = sum(w(:).*phip(u(:),p,S0));    